/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * matrix.c
 *
 * This module hides the details of stringpool and matrix handling.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "calc.h"
#include "types.h"
#include "../xstuff/fonts.h"

int max_lines = BUFFER_ROWS;
int max_columns = BUFFER_COLS;

#define POOLSIZE 100000
static char *stringpool;

/*X
   spread **new_matrix()
   Creates new empty matrix.  Returns NULL if it failed.
   X */
spread **new_matrix()
{
	int i;
	spread **matrix;

	if ((matrix = (spread **)calloc(max_lines + 1, sizeof(spread *)))
			== NULL)
		return NULL;

	for (i = 0; i < max_lines + 1; matrix[i++] = NULL);
	return matrix;
}

/*X
   void free_matrix(spread **matrix)
   Frees the memory used by the matrix.
   X */
void free_matrix(spread ** matrix)
{
	int i;

	for (i = 0; i < max_lines + 1; i++)
		if (matrix[i] != NULL)
			free((char *) matrix[i]);
	free((char *) matrix);
}

int init_pool()
{
	if ((stringpool = (char *) malloc(POOLSIZE + 1)) == NULL)
		return 1;
	stringpool[1] = '\0';	/* stringpool[0] not used; index 0 = error */
	return 0;
}

short pool_size()
{
	register char *pi;
	register int n;

	for (pi = stringpool + 1; *pi; pi += (n + 1)) {
		n = *pi++;
	}
	return pi - stringpool;
}

/* I'm sick of all the different stpcpy's so here goes */
static char *mystpcpy(char *to, char *from)
{
	strcpy(to, from);
	return to+strlen(to);
}

short pool_install(char *p)
/* Returns: 0 if error, otherwise index of installed string */
{
/*  register short i; */
	register char *pi;	/* faster than stringpool[i] */
	register int m, n;

	m = strlen(p);
	for (pi = stringpool + 1; *pi; pi += (n + 1)) {
		n = *pi++;
		if (m == n && !strcmp(p, pi))
			return pi - stringpool;
	}
	if (pi - stringpool > POOLSIZE)
		return 0;

	*pi++ = m;
	*(mystpcpy(pi, p) + 1) = '\0';	/* extra null byte marks the end */
	return pi - stringpool;
}				/* pool_install */

int ins_data(spread ** matrix, short texti, double v, short t,
	     int row, int col)
{
	int i, change = FALSE;
	register spread *mrowcol;	/* faster than matrix[row][col] */

	if (matrix[row] == NULL) {
		if (t == EMPTY)
			return FALSE;
		else {
			matrix[row] = (spread *)calloc(max_columns + 1, sizeof(spread));
			for (i = 1; i < max_columns + 1; i++)
				matrix[row][i].format = SIZE_10 | HELVETICA;
		}
		/* error check should be in here somewhere */
	}
	mrowcol = matrix[row] + col;
	if (mrowcol->text != texti || mrowcol->type != t)
		change = TRUE;
	mrowcol->text = texti;
	mrowcol->value = v;
	mrowcol->type = t;
	return change;
}				/* ins_data */

int ins_format(spread ** matrix, int row, int col, int format)
{
	int i;

	if (matrix[row] == NULL) {
		matrix[row] = (spread *) calloc(max_columns + 1, sizeof(spread));
		for (i = 0; i < max_columns - 1; i++)
			matrix[row][i].format = SIZE_10 | HELVETICA;
	}
	matrix[row][col].format = format;
	return 1;
}

short ret_text(spread ** matrix, int row, int col)
{
	register spread *mrow;	/* faster than matrix[row] */
	if ((mrow = matrix[row]))
		return mrow[col].text;
	else
		return 0;
}				/* ret_text */

char *
ret_taddr(spread ** matrix, int row, int col)
{
	int i;
	if ((i = ret_text(matrix, row, col)))
		return stringpool + i;
	else
		return NULL;
}

double ret_val(spread ** matrix, int row, int col)
{
	register spread *mrow;	/* faster than matrix[row] */
	if ((mrow = matrix[row]))
		return mrow[col].value;
	else
		return 0;
}				/* ret_val */

short ret_type(spread ** matrix, int row, int col)
{
	register spread *mrow;	/* faster than matrix[row] */
	if ((mrow = matrix[row]))
		return mrow[col].type;
	else
		return EMPTY;
}				/* ret_type */

int ret_format(spread ** matrix, int row, int col)
{
	if ((matrix[row]))
		return matrix[row][col].format;
	else
		return SIZE_10 | HELVETICA;
}

int ret_font(spread ** matrix, int row, int col)
{
	return ret_format(matrix, row, col) & 127;
}

int line_last_used(spread ** matrix)
{
	register int r;
	for (r = max_lines; r > 1 && matrix[r] == NULL; r--);
	return r;
}

int col_last_used(spread ** matrix, int row)
{
	register int i;
	for (i = max_columns; i > 1 && ret_type(matrix, row, i) == EMPTY; i--);
	return i;
}

void downshift_matrix(buffer * b, int row)
{
	int i;
	if (b->matrix[BUFFER_ROWS])
		free(b->matrix[BUFFER_ROWS]);
	for (i = BUFFER_ROWS; i > row; i--) {
		b->matrix[i] = b->matrix[i - 1];
		b->height[i] = b->height[i - 1];
	}
	b->matrix[row] = NULL;
}

void upshift_matrix(buffer * b, int row)
{
	int i;
	if (b->matrix[row])
		free(b->matrix[row]);
	for (i = row; i < BUFFER_ROWS; i++) {
		b->matrix[i] = b->matrix[i + 1];
		b->height[i] = b->height[i + 1];
	}
	b->matrix[BUFFER_ROWS] = NULL;
	b->height[BUFFER_ROWS] = 20;
}

void rightshift_matrix(buffer * b, int col)
{
	int i, j;
	for (j = BUFFER_COLS; j > col; j--)
		b->width[j] = b->width[j - 1];
	for (i = 1; i <= BUFFER_ROWS; i++) {
		if (b->matrix[i]) {
			for (j = BUFFER_COLS; j > col; j--) {
				b->matrix[i][j] = b->matrix[i][j - 1];
			}
			b->matrix[i][col].type = EMPTY;
		}
	}
}

void leftshift_matrix(buffer * b, int col)
{
	int i, j;
	for (j = col; j < BUFFER_COLS; j++)
		b->width[j] = b->width[j + 1];
	for (i = 1; i <= BUFFER_ROWS; i++) {
		if (b->matrix[i]) {
			for (j = col; j < BUFFER_COLS; j++) {
				b->matrix[i][j] = b->matrix[i][j + 1];
			}
			b->matrix[i][BUFFER_COLS].type = EMPTY;
		}
	}
}
